# usage.py

"""
Options and arguments to bo passed on to the program.
"""

import os
import sys
import argparse
import info


class EditConfig(argparse.Action):
    """
    https://stackoverflow.com/questions/8632354/python-argparse-custom-actions-with-additional-arguments-passed
    https://stackoverflow.com/questions/11001678/argparse-custom-action-with-no-argument
    """
    def __init__(self, option_strings, conf, *args, **kwargs):
        """ Function doc """
        super(EditConfig, self).__init__(option_strings=option_strings, *args, **kwargs)
        self.conf = conf

    def __call__(self, parser, namespace, values, option_strings):
        self.conf.edit()
        sys.exit()


def usage(conf):
    """
        Help function with argument parser.
    """
    parser = argparse.ArgumentParser(
        description=info.DOC,
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    query_type = parser.add_mutually_exclusive_group(required=True)
    query_type.add_argument('-s', '--selection',
                            help=(
                                "list of gene IDs or transcript IDs (ENST, ENS or gene Symbol) to "
                                "select inside your fasta transcriptome file and that you want to "
                                "extract specific kmers from. For genes, kmerator search specific "
                                " kmers along the gene. For transcripts, it search specific kmers "
                                "to the transcript. You can also give a file with yours "
                                "genes/transcripts separated by space, tab or newline. "
                                "If you want to use your own unannotated sequences, you "
                                "must give your fasta file with --fasta_file option."
                                ),
                            nargs='+',
                            )
    query_type.add_argument('-f', '--fasta-file',
                            help=(
                                "Use this option when yours sequences are unannonated or provided "
                                "by a annotation file external from Ensembl. Otherwise, "
                                "use --selection option."
                                ),
                            )
    parser.add_argument('-g', '--genome',
                        help=(
                            "genome fasta file or jellyfish index (.jf) to use for k-mers requests."
                            ),
                        required=True,
                        )
    parser.add_argument('-t', '--transcriptome',
                        help=(
                            "transcriptome fasta file (ENSEMBL fasta format ONLY) to use for "
                            "k-mers request and transcriptional variants informations."
                            ),
                        required=True,
                        )
    parser.add_argument('-j', '--jellyfish-transcriptome',
                        help=(
                            "if your transcriptome (-t option) has already been converted by "
                            "jellyfish as a jf file, this avoids redoing the operation (be careful,"
                            "it must be the same transcriptome!)."
                            ),
                        )
    parser.add_argument('-S', '--specie',  ### replace -a --appris in julia version
                        help=(
                            "indicate a specie referenced in Ensembl, to help, follow the link "
                            "https://rest.ensembl.org/documentation/info/species. You can use "
                            "the 'name', the 'display_name' or any 'aliases'. For example human, "
                            "homo_sapiens or homsap are valid (default: human)."
                            ),
                        default='human',
                        )
    parser.add_argument('-k', '--kmer-length',
                        type=int,
                        help="k-mer length that you want to use (default 31).",
                        default=31,
                        )
    parser.add_argument('--chimera',
                        action='store_true',
                        help="Only if with --fasta-file option.",
                        )
    parser.add_argument('--stringent',
                        action='store_true',
                        help=(
                            "Only for genes with --selection option: use this option if you want"
                            "to select gene-specific k-mers present in ALL known transcripts for"
                            "your gene. If false, a k-mer is considered as gene-specific if present"
                            "in at least one isoform of your gene of interest."
                            ),
                        )
    parser.add_argument('--max-on-transcriptome',
                        type=float,
                        help=argparse.SUPPRESS,
                        default=0,
                        )
    parser.add_argument('-o', '--output',
                        help="output directory (default: 'output')",
                        default='output',
                        )
    parser.add_argument('-p', '--procs',
                        type=int,
                        help="run n process simultaneously (default: 1)",
                        default=1,
                        )
    parser.add_argument('-d', '--debug',
                        action='store_true',
                        help="if you want some details while Kmerator is running.",
                        )
    parser.add_argument('--keep',
                        action='store_true',
                        help=("keep intermediate files (sequences, indexes, separate tags and "
                            "contigs files)."
                            ),
                        )
    parser.add_argument('-e', '--edit-config',
                        action=EditConfig,
                        nargs=False,
                        conf=conf,
                        # ~ metavar=('edit config file'),
                        help='Edit config file',
                        )
    parser.add_argument('-v', '--version',
                        action='version',
                        version=f'{parser.prog} v{info.VERSION}',
                        )

    ### Set default values define by Config class (using ConfigParser)
    default_conf = conf.content['CMD_ARGS']
    parser.set_defaults(**default_conf)
    ### Reset `required` attribute when provided from config file
    for action in parser._actions:
        if action.dest in default_conf:
            action.required = False
            ### handle booleans cause ConfigParse return a string.
            # ~ print(action)
            if action.__class__.__name__ == "_StoreTrueAction":
                action.default = True if action.default.lower() in ['true', 'yes', '1'] else False

    ### Go to 'usage()' without arguments or stdin
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit()
    args = parser.parse_args()

    return args


def checkup_args(args):
    """ test genome and transcriptome files. """
    ### check if query fasta file is present
    if args.fasta_file and not os.path.isfile(args.fasta_file):
        sys.exit(f"{Color.RED}Error: {args.fasta_file!r} not found{Color.END}.")
    ### Check query fasta file extension
    if args.fasta_file and args.fasta_file.split('.')[-1] not in ("fa", "fasta"):
        sys.exit(f" {Color.RED}Error: {os.path.basename(args.fasta_file)} " \
                f"Does not appears to be a fasta file.{Color.END}")
    ### check if genome is present
    if not os.path.isfile(args.genome):
        sys.exit(f"{Color.RED}Error: {args.genome!r} not found{Color.END}.")
    ### Check genome fasta/jellyfish file extension
    if args.genome.split('.')[-1] not in ("fa", "fasta", "fna", "jf", "jl"):
        sys.exit(f" {Color.RED}Error: {os.path.basename(args.genome.name)}" \
                f"Does not appears to be a fasta or jelifish file.{Color.END}")
    ### check if transcriptome is present
    if args.transcriptome and not os.path.isfile(args.transcriptome):
        sys.exit(f"{Color.RED}Error: {args.transcriptome!r} not found{Color.END}.")
    ### Check Gene Select option
    if args.selection:
        for gene in args.selection:
            if gene.startswith('ENS') and '.' in gene:
                sys.exit(f"{Color.RED}ENSEMBL annotations with version (point) like "
                        f"ENSTXXXXXXXX.1 is forbidden, just remove the '.1'.{Color.END}"
                        )
        ### Define genes/transcripts provided when they are in a file
        if len(args.selection) == 1 and os.path.isfile(args.selection[0]):
            with open(args.selection[0]) as fh:
                args.selection = fh.read().split()
    ## --chimera level works only with --fasta-file option
    if args.chimera and not args.fasta_file:
        sys.exit(f"{Color.RED}Error: '--chimera' needs '--fasta-file' option.{Color.END}")
    ### if jellyfish of transcriptome is provided, check it
    if args.jellyfish_transcriptome and not args.jellyfish_transcriptome.split('.')[-1] == 'jf':
        sys.exit(f"{Color.RED}Error: {os.path.basename(args.jellyfish_transcriptome)!r} does not seem to be in jellyfish format.")


'''
def is_transcriptome_ensembl_file(first_row):
    """Checks if the line matches the Ensembl transcriptome file format"""
    line = first_row.split()
    try:
        if (not line[6].startswith('gene_symbol:') or
            not line[0].startswith('>ENST') or
            not line[3].startswith('gene:')):
            return False
    except:
        return False
    return True
'''

class Color:
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'



#!/usr/bin/env python3

import sys
import os
import requests
import collections
from concurrent.futures import ThreadPoolExecutor
import time
from utils import Color
import json


def main():
    """ Function doc """
    ### some variables
    report = {'aborted': [], 'done': [], 'multiple': []}

    ### get transcript from Ensembl API
    print("GET ENSEMBL INFO")
    ensembl = Ensembl(args, report)
    transcripts = ensembl.get_ENST()

    print(f"Selection ({len(args.selection)})")
    print(*args.selection)

    print(f"\ntranscripts found ({len(transcripts)})")
    for k,v in transcripts.items():
        print(f"  {k}: {v}")

    print("\nREPORT")
    print(' done:', report['done'])
    print(' aborted:', report['aborted'])
    print(' multiple:', report['multiple'])


class Ensembl:
    """ Class doc """
    url = "https://rest.ensembl.org"
    nb_threads = 15                     # Because Ensembl limits requests to 15 per second'

    def __init__(self, args, transcript2gene_dict, report):
        """ Class initialiser """
        self.transcript2gene_dict = transcript2gene_dict
        self.report = report
        self.args = args
        self.headers = {"Content-Type" : "application/json"}   # header for the query


    def get_ENST(self):
        """ For each gene-symbol, ENSG or ENST,  """
        transcripts = {}                # the dict to return
        urls = []                       # list of dicts where keys are 'item', 'level' and 'url'
        symbols_urls = []

        ### Define Ensembl URLs for each given item
        for item in self.args.selection:
            if item.upper().startswith('ENST'):
                urls.append({'level': 'transcript',
                             'item': item,
                             'url': f"{Ensembl.url}/lookup/id/{item}?",
                            })
            elif item.upper().startswith('ENSG'):
                urls.append({'level': 'gene',
                             'item': item,
                             'url': f"{Ensembl.url}/lookup/id/{item}?",
                            })
            else:
                symbols_urls.append({'item': item,
                                    'url': f"{Ensembl.url}/xrefs/symbol/{self.args.specie}/{item}?"})

        ### fetch ENSGxxx from each GENE SYMBOL
        with ThreadPoolExecutor(Ensembl.nb_threads) as pool:
            ensg_symbol_list = list(pool.map(self.ebl_request, symbols_urls))

        ### reduce symbol list to ENSG
        for symbol in ensg_symbol_list:
            symbol['response'] = [a['id'] for a in symbol['response'] if a['id'].startswith('ENSG')]

        ### Define symbol urls for urls list
        for symbol in ensg_symbol_list:
            if len(symbol['response']) == 0:
                # ~ self.report['aborted'].append(f"{symbol['item']!r} not found by Ensembl API.")
                continue
            for ensg in symbol['response']:
                item = symbol['item']
                urls.append({'level': 'gene',
                              'item': item,
                              'url': f"{Ensembl.url}/lookup/id/{ensg}?",
                            })

        ### fetch Ensembl ENSTxxx matching with given GENES, ENSGxxx and ENSTxxx
        find_xtime = collections.defaultdict(list)
        with ThreadPoolExecutor(Ensembl.nb_threads) as pool:
            enst_list = list(pool.map(self.ebl_request, urls))
        for entry in enst_list:
            response = entry['response']
            if not response:
                continue
            level = entry['level']
            item = entry['item']
            # ~ if  not 'display_name' in response:
                # ~ self.report['aborted'].append(f"{item!r} not found by Ensembl API.")
                # ~ continue
            if response['seq_region_name'].startswith('CHR_'):
                if not entry['item'] in find_xtime:  # to avoid repetition when multiples ENST are found for once gene
                    self.report['aborted'].append(f"{item!r} found in an alternative chromosome (Ensembl API).")
                continue
            if level == 'transcript':
                transcript = response['id'].upper()
            else:
                transcript = response['canonical_transcript'].split('.')[0]
            ### create transcript dict
            if transcript in self.transcript2gene_dict:
                symbol = self.transcript2gene_dict[transcript]
                transcripts[transcript] = {'given': item, 'symbol': symbol, 'level': level}
                if level == 'gene':
                    find_xtime[item].append(symbol)
            else:
                ### If transcript not found in transcriptome (transcript from Ensembl info, but not same local transcriptome version)
                self.report['aborted'].append(f"{transcript!r} not found in transcriptome ({item}).")

        ### add multiple ENSTxxx found per gene
        for item, names in find_xtime.items():
            if len(names) > 1:
                self.report['multiple'].append({item: names})

        return transcripts


    def ebl_request(self, query):
        start_time = time.time()                # Because Ensembl limits requests to 15 per second
        ### part of query is returned in response
        response = query
        try:
            r = requests.get(query['url'], headers=self.headers)
        except requests.ConnectionError as err:
            sys.exit(f"{Color.RED}\n Error: Ensembl is not accessible or not responding.{Color.END}")
        ### remove useless info
        response.pop('url')
        if not r.ok:
            # ~ self.report['aborted'].append(f"{query['item']!r}: something went wrong whis the request to Ensembl.")
            r = []
        else:
            r = r.json()
        if not r:
            self.report['aborted'].append(f"{query['item']!r} not found by the Ensembl API (or error).")
            r = []
        elif 'error' in r:
            self.report['aborted'].append(f"{query['item']!r} not found by Ensembl API.")
            r = []
        ### Because Ensembl limits requests to 15 per second
        limit = max(0, 1-(time.time()-start_time)+0.05)
        time.sleep(limit)
        response['response'] = r

        return response


    def _find_longest_variant(self, entry):
        print("entry:", entry)
        return
        if args.verbose: print(f"{'-'*9}\n{Color.YELLOW}Finding the longest variant for the gene {entry[item]}.{Color.END}")
        variants_dict = { k:len(v) for (k,v) in transcriptome_dict.items() if k.startswith(f"{gene_name}:")}
        # ~ print(*[k for k in variants_dict], sep='\n')
        nb_variants = len(variants_dict)
        if args.verbose: print(f"{Color.YELLOW}Number of variants: {nb_variants}")
        longest_variant = None
        length = 0
        for k,v in variants_dict.items():
            if v > length:
                length = v
                longest_variant = ':'.join(k.split(':')[1:2])
        # ~ print(f"{longest_variant = }")
        return longest_variant


class args:
    """ Class doc """
    specie = 'human'
    # ~ selection = [ 'GLOP', 'BRCA2', 'NPM1', 'CHI3L1', 'VPS29', 'ENST00000621131', 'ENSG00000159216',
        # ~ 'ENST00000614774', 'hstf1', 'braf', 'tp53', 'kras', 'egfr', 'ret', 'met', 'alk',
        # ~ 'BRCA1', 'stk11', 'ros1', 'notch1', 'CD46', 'GNG11', 'ADAMTS4', 'PCDH9', 'ZNF703', 'ZNF462'
        # ~ 'ZNF569', 'SNORA23', 'TXNRD1', 'ERRFI1', 'MARVELD2', 'DEPTOR', 'ELN', 'ADAMTS7', 'NOTCH3'
    # ~ ]
    selection = ['hstf1', 'glOp', 'enst00000614774', 'enst00000621131', 'enst00006211431', 'npm1', 'ensg00000159216']
    # ~ selection = ['hstf1']


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""
Decomposition of transcript or gene sequences and extraction of their specific k-mers
1. Load transcriptome
2. Request Ensembl - build best transcripts
3. Build sequences for each transcript
4. Get specific kmers
"""


import sys
import os
import requests
import shutil
import getpass
from datetime import datetime
import signal
from functools import partial

import info
import sequences
from sequences import Sequences
from utils import usage, checkup_args, Color
from kmerize import SpecificKmers
from ensembl import Ensembl
from config import Config



def main():
    ### Handle arguments
    conf = Config(info.APPNAME)
    args = usage(conf)
    if args.debug: print(f"{'-'*9}\n{Color.YELLOW}Args: {args}{Color.END}")

    ### Handle control C
    global sigint_handler
    sigint_handler = partial(sigint_handler, args=args)
    signal.signal(signal.SIGINT, sigint_handler)

    ### check options
    checkup_args(args)

    ### some variables
    report = {'aborted': [], 'done': [], 'multiple': [], 'warning': []}
    transcriptome_dict = {}
    transcript2gene_dict = {}
    best_transcripts = {}                       # when genes/transcripts annotated (--selection)
    unannotated_transcripts = []                # when transcripts are unannotated (--fasta-file)

    ### when --selection option is set
    if args.selection:
        ### Load transcriptome as dict (needed to build sequences and to found specific kmers)
        print(f" 🧬 Load transcriptome.")
        transcriptome_dict, transcript2gene_dict = ebl_fasta2dict(args.transcriptome)
        ### get canonical transcripts using Ensembl API
        print(f" 🧬 Fetch some information from Ensembl API.")
        ensembl = Ensembl(args, transcript2gene_dict, report)
        best_transcripts = ensembl.get_ENST()
        ### Build sequence using provided transcriptome
        print(f" 🧬 Build sequences.")
        seq = Sequences(args, report, best_transcripts, transcriptome_dict)
        seq.build()
    ### when --fasta-file option is set
    else:
        print(f" 🧬 Build sequences.")
        # ~ sequences.build(args, report, unannotated_transcripts)
        seq = Sequences(args, report, unannotated_transcripts)
        seq.build()

    ### get specific kmer with multithreading
    print(f" 🧬 Extract specific kmers, please wait..")
    kmers = SpecificKmers(args, report, transcriptome_dict, best_transcripts, unannotated_transcripts)

    ### Concatene results
    merged_results(args)

    ### show some final info in the prompt
    show_info(report)

    ### set markdown report
    markdown_report(args, report)

    ### ending
    exit_gracefully(args)


def ebl_request(report, item, url, headers):
    try:
        r = requests.get(url, headers=headers)
    except requests.ConnectionError as err:
        sys.exit(f"{Color.RED}\n Error: Ensembl is not accessible or not responding.{Color.END}")
    r = r.json()
    if not r:
        report['aborted'].append(f"{item} not found from Ensembl API.")
        return None
    if 'error' in r:
        report['aborted'].append(f"{r[error]}.")
        return None
    return r


def ebl_fasta2dict(fasta_file):
    '''
    Convertion from Ensembl fasta file to dict
    It keeps only the transcript name of the headers, without version number
    '''
    ### controls
    with open(fasta_file) as fh:
        first_line = fh.readline()
        if not first_line.startswith('>'):
            sys.exit(f"{Color.RED}Error: {os.path.basename(args.fasta_file.name)!r} does not seem to be in fasta format.")
    ### compute file as dict
    fasta_dict = {}
    transcript2gene_dict = {}
    with open(fasta_file) as fh:
        seq = ""
        old_desc, new_desc = "", ""
        for line in fh:
            if line[0] == ">":
                line = line.split()
                if len(line) > 6 and line[6].split(':')[0] == 'gene_symbol':
                    gene_name = line[6].split(':')[1]                    # gene symbol
                else:
                    gene_name = line[3].split(':')[1].split('.')[0]      # ENSG
                transcript_name = line[0].split('.')[0].lstrip('>')
                new_desc = f"{gene_name}:{transcript_name}"
                transcript2gene_dict[transcript_name] = gene_name
                # ~ new_desc = transcript_name
                if old_desc:
                    fasta_dict[old_desc] = seq
                    seq = ""
                old_desc = new_desc
            else:
                seq += line.rstrip()
    fasta_dict[old_desc] = seq
    return fasta_dict, transcript2gene_dict


def merged_results(args):
    if not os.path.isdir(os.path.join(args.output, 'tags')):
        return None
    for item in ['tags', 'contigs']:
        files = os.listdir(os.path.join(args.output, item))
        if files:
            merged_file = os.path.join(args.output, f"{item}-merged.fa")
            with open(merged_file,'wb') as mergefd:
                for file in files:
                    with open(os.path.join(args.output, item, file),'rb') as fd:
                        shutil.copyfileobj(fd, mergefd)


def show_info(report):
    ### show some final info in the prompt
    print(f"{Color.CYAN}\n Done ({len(report['done'])}):")
    for i,mesg in enumerate(report['done']):
        if i == 15:
            print("  - ... (more responses)")
            break
        print(f"  - {mesg}")


    if report['multiple']:
        print(f"{Color.BLUE}\n Multiple responses from Ensembl API ({len(report['multiple'])}):")
        for i,mesg in enumerate(report['multiple']):
            if i == 15:
                print("  - ... (more responses)")
                break
            for k,v in mesg.items():
                print(f"  - {k}: {' '.join(v)}")

    if report['aborted']:
        print(f"{Color.PURPLE}\n Aborted ({len(report['aborted'])}):")
        for i,mesg in enumerate(report['aborted']):
            if i == 15:
                print("  - ... (more responses)")
                break
            print(f"  - {mesg}")

    if report['warning']:
        print(f"{Color.RED}\n Warning ({len(report['warning'])}):")
        for i,mesg in enumerate(report['warning']):
            if i == 15:
                print("  - ... (more responses)")
                break
            print(f"  - {mesg}")



    print(f"{Color.END}")


def markdown_report(args, report):
    with open(os.path.join(args.output, 'report.md'), 'w') as fh:
        fh.write('# kmerator report\n')
        fh.write(f"*date: {datetime.now().strftime('%Y-%m-%d %H:%M')}*  \n")
        fh.write(f'*login: {getpass.getuser()}*\n\n')
        fh.write(f"**kmerator version:** {info.VERSION}\n\n")
        ### report command line and args, included defaults args
        cmd_args = ''
        for k,v in vars(args).items():
            if v and k != 'jellyfish_genome':
                k = k.replace('_', '-')
                if isinstance(v, list):
                    v = ' '.join(v)
                cmd_args += f" \\\n  --{k} {v}"
        # ~ cmd_args = ' \\\n  '.join([f"--{k} {v}" for k,v in vars(args).items() if v])
        # ~ command = f"command: \n{__appname__}{cmd_args}"
        command = f"{info.APPNAME}{cmd_args}"
        # ~ command = ' '.join(sys.argv).replace(' -', ' \\\n  -')
        fh.write(f"**Command:**\n\n```\n{command}\n```\n\n")
        fh.write(f"**Working directory:** `{os.getcwd()}`\n\n")
        fh.write(f"**Jellyfish transcriptome used:** `{args.jellyfish_transcriptome}`\n\n")
        if report['done']:
            fh.write(f"**Genes/transcripts succesfully done ({len(report['done'])})**\n\n")
            for mesg in report['done']:
                fh.write(f"- {mesg}\n")
        if report['multiple']:
            fh.write(f"\n**Multiple Genes returned for one given by Ensembl API ({len(report['multiple'])})**\n\n")
            for mesg in report['multiple']:
                for k,v in mesg.items():
                    fh.write(f"- {k}: {' '.join(v)}\n")
        if report['aborted']:
            fh.write(f"\n\n**Genes/transcript missing ({len(report['aborted'])})**\n\n")
            for mesg in report['aborted']:
                fh.write(f"- {mesg}\n")
        if report['warning']:
            fh.write(f"\n\n**Warnings ({len(report['warning'])})**\n\n")
            for mesg in report['warning']:
                fh.write(f"- {mesg}\n")


def sigint_handler(signal, frame, args):
    exit_gracefully(args)


def exit_gracefully(args):
    if not args.keep:
        shutil.rmtree(os.path.join(args.output,'tags'), ignore_errors=True)
        shutil.rmtree(os.path.join(args.output,'contigs'), ignore_errors=True)
        shutil.rmtree(os.path.join(args.output,'indexes'), ignore_errors=True)
        shutil.rmtree(os.path.join(args.output,'sequences'), ignore_errors=True)
    sys.exit(0)


""" _find_longest_variant
def _find_longest_variant(args, gene_name, transcriptome_dict):
    if args.verbose: print(f"{'-'*9}\n{Color.YELLOW}Finding the longest variant for the gene {gene_name}.{Color.END}")
    variants_dict = { k:len(v) for (k,v) in transcriptome_dict.items() if k.startswith(f"{gene_name}:")}
    # ~ print(*[k for k in variants_dict], sep='\n')
    nb_variants = len(variants_dict)
    if args.verbose: print(f"{Color.YELLOW}Number of variants: {nb_variants}")
    longest_variant = None
    length = 0
    for k,v in variants_dict.items():
        if v > length:
            length = v
            longest_variant = ':'.join(k.split(':')[1:2])
    # ~ print(f"{longest_variant = }")
    return longest_variant
"""


if __name__ == '__main__':
    main()



### kmerize.py

import os
import sys
import subprocess
import multiprocessing

from fasta import fasta2dict
from utils import Color


class SpecificKmers:
    """ Class doc """

    def __init__(self, args, report, transcriptome_dict, best_transcripts, unannotated_transcripts):
        """ Class initialiser """
        self.args = args
        self.rev = rev = {'A':'T', 'C':'G', 'G':'C', 'T':'A',
                          'a':'t', 'c':'g', 'g':'c', 't':'a'}       # reverse base
        ### create a shared dict among multiple processes with Manager()
        ### (show https://ourpython.com/python/multiprocessing-how-do-i-share-a-dict-among-multiple-processes)
        manager = multiprocessing.Manager()
        self.transcriptome_dict = manager.dict(transcriptome_dict)
        self.best_transcripts = manager.dict(best_transcripts)
        ### compute Jellyfish on genome and transcriptome if not exists
        self.jellyfish()
        ### Sequences files to analyse
        self.seq_files_dir = os.path.join(self.args.output, 'sequences')
        ### launch workers
        transcripts = best_transcripts.items() if args.selection else unannotated_transcripts
        with multiprocessing.Pool(processes=self.args.procs) as pool:
            messages = pool.map(self.worker, transcripts)
        for type,mesg in messages:
            report[type].append(mesg)


    def worker(self, transcript):
        '''
        transcript is a dict when '--selection' is set, else it is a list
        '''
        ### Define some variables: gene_name, transcript_name, variants_dic and output file names
        fasta_kmer_list = []                # specific kmers list
        fasta_contig_list = []              # specific contigs list
        ## When '--selection' option is set
        if self.args.selection:
            transcript_name = transcript[0]          # ENST00000001
            gene_name = transcript[1]['symbol']      # tp53
            level = transcript[1]['level']           # 'gene' or 'transcript'
            given_name = transcript[1]['given']      # TP53 (gene/transcript given by user)
            seq_file = f"{gene_name}.{transcript_name}.fa"
            ### Define all variants for a gene
            variants_dict = { k:v for k,v in self.transcriptome_dict.items() if k.split(":")[0] == gene_name }
            nb_isoforms = len(variants_dict)
            tag_file = f"{gene_name}-{transcript_name}-{level}-specific_kmers.fa"
            contig_file = f"{gene_name}-{transcript_name}-{level}-specific_contigs.fa"
        ## When '--chimera' option is set
        elif self.args.chimera:
            seq_file = f"{transcript.replace(' ', '_').replace('/', '@SLASH@')}.fa"[:255]
            gene_name = transcript_name = transcript
            level = 'chimera'
            tag_file = f"{gene_name}-chimera-specific_kmers.fa"
            contig_file = f"{gene_name}-chimera-specific_contigs.fa"
        ## When '--fasta-file' option is set
        else:
            seq_file = f"{transcript.replace(' ', '_').replace('/', '@SLASH@')}.fa"[:255]
            gene_name = transcript_name = transcript
            level = 'transcript'
            tag_file = f"{gene_name}-transcript-specific_kmers.fa"
            contig_file = f"{gene_name}-transcript-specific_contigs.fa"

        ### take the transcript sequence for jellyfish query
        sequence_fasta = fasta2dict(os.path.join(self.args.output,'sequences', seq_file))

        ### building kmercounts dictionary using jellyfish query on the genome
        cmd = (f"jellyfish query -s '{os.path.join(self.seq_files_dir,seq_file)}' {self.args.jellyfish_genome}")
        try:
            kmercounts_genome = subprocess.run(cmd, shell=True, check=True, capture_output=True).stdout.decode().rstrip().split('\n')
        except subprocess.CalledProcessError:
            sys.exit(f"{Color.RED}Error: an error occured in jellyfish query command for {seq_file}:\n  {cmd}{Color.END}")
            return None
        kmercounts_genome_dict = {}
        for mer in kmercounts_genome:
            seq, count = mer.split()
            kmercounts_genome_dict[seq] = int(count)

        ### building kmercounts dictionary using jellyfish query on the transcriptome
        cmd = (f"jellyfish query -s '{os.path.join(self.seq_files_dir,seq_file)}' {self.args.jellyfish_transcriptome}")
        try:
            kmercounts_transcriptome = subprocess.run(cmd, shell=True, check=True, capture_output=True).stdout.decode().rstrip().split('\n')
        except subprocess.CalledProcessError:
            self.report['warning'].append(f"an error occured in jellyfish query command for {seq_file}:\n  {cmd}")
            return None
        kmercounts_transcriptome_dict = {}
        for mer in kmercounts_transcriptome:
            seq, count = mer.split()
            kmercounts_transcriptome_dict[seq] = int(count)

        ### initialization of count variables
        i = 0       # kmer number (selected kmer)
        j = 1       # contig number
        total_kmers = len(kmercounts_transcriptome_dict)
        if self.args.debug: print(f"{Color.YELLOW}kmer counts found by Jellyfish for {seq_file}: {total_kmers}{Color.END}")

        ## creating a new dictionary with kmers and their first position in our query sequence
        kmer_starts = {}
        kmer_placed = 0

        for mer in kmercounts_transcriptome_dict:
            ### get the first position of the kmer in the sequence
            kmer_placed += 1
            kmer_starts[mer] = next(iter(sequence_fasta.values())).index(mer)

        ### rearrange kmer_starts as list of sorted tuple like (position, kmer)
        kmer_starts_sorted = sorted(list(zip(kmer_starts.values(), kmer_starts.keys())))  # array sorted by kmer position
        # ~ position_kmer_prev = first(kmer_starts_sorted[1])
        position_kmer_prev = kmer_starts_sorted[0][0]
        contig_seq = "" # initialize contig sequence
        ### for each kmer, get the count in both genome and transcriptome
        for tuple in kmer_starts_sorted:
            ### from the kmer/position sorted list, we extract sequence if specific (occurence ==1)
            position_kmer = tuple[0]    # position of kmer
            mer = tuple[1]              # sequence of kmer
            kmers_analysed = position_kmer+1
            per = round(kmers_analysed/total_kmers*100)     # to show percentage done ?

            if mer in kmercounts_genome_dict.keys():
                genome_count = kmercounts_genome_dict[mer]
            else:
                revcomp_mer = ''.join([self.rev[base] for base in mer][::-1])
                genome_count = kmercounts_genome_dict[revcomp_mer]
            transcriptome_count = kmercounts_transcriptome_dict[mer]
            ### Case of annotated genes
            if level == 'gene':
                ### if the kmer is present/unique or does not exist (splicing?) on the genome
                if genome_count <= 1:
                    ### who are the variants of this kmer
                    isoforms_containing_this_kmer = [k for k,v in variants_dict.items() if mer in v]
                    nb_iso = len(isoforms_containing_this_kmer)
                    if  (self.args.stringent
                            and transcriptome_count == nb_isoforms == len(isoforms_containing_this_kmer)
                        ):
                        # kmers case
                        i += 1
                        fasta_kmer_list.append(f">{gene_name}:{transcript_name}.kmer{kmers_analysed} ({nb_iso}/{nb_isoforms})\n{mer}")
                        # contigs case
                        if i == 1:
                            contig_seq = mer
                            position_kmer_prev = position_kmer
                        elif i>1 and position_kmer == position_kmer_prev+1:
                            contig_seq = f"{contig_seq}{mer[-1]}"
                            position_kmer_prev = position_kmer
                        else:
                            fasta_contig_list.append(f">{gene_name}:{transcript_name}.contig{j}\n{contig_seq}")
                            j = j+1
                            contig_seq = mer
                            position_kmer_prev = position_kmer
                    elif (not self.args.stringent
                            and transcriptome_count == len(isoforms_containing_this_kmer)
                            and transcriptome_count > 0
                         ):
                        ### kmers case
                        i += 1
                        fasta_kmer_list.append(f">{gene_name}:{transcript_name}.kmer{kmers_analysed} ({nb_iso}/{nb_isoforms})\n{mer}")
                        ### contigs case
                        if i == 1:
                            contig_seq = mer
                            position_kmer_prev = position_kmer
                        elif i > 1 and position_kmer == position_kmer_prev + 1:
                            contig_seq = f"{contig_seq}{mer[-1]}"
                            position_kmer_prev = position_kmer
                        else:
                            fasta_contig_list.append(f">{gene_name}:{transcript_name}.contig{j}\n{contig_seq}")
                            j += 1
                            contig_seq = mer
                            position_kmer_prev = position_kmer

            ### Cases of transcripts 1) unannotated, 2) annotated.
            elif level == 'transcript':
                ### Case of unannotated transcripts
                if self.args.fasta_file and transcriptome_count <= self.args.max_on_transcriptome and genome_count <= 1: # max_on_transcriptome = 0 by default
                    ### kmers case
                    i += 1
                    fasta_kmer_list.append(f">{gene_name}.kmer{kmers_analysed}\n{mer}")
                    ### contigs case
                    if i == 1:
                        contig_seq = mer
                        position_kmer_prev = position_kmer
                    elif i > 1 and position_kmer == position_kmer_prev+1:
                        contig_seq = f"{contig_seq}{mer[-1]}"
                        position_kmer_prev = position_kmer
                    else:
                        fasta_contig_list.append(f">{gene_name}.contig{j}\n{contig_seq}")
                        j += 1
                        contig_seq = mer
                        position_kmer_prev = position_kmer
                ### Case of annotated transcripts
                elif self.args.selection and transcriptome_count == 1 and genome_count <= 1:
                    ### kmers case
                    i += 1
                    fasta_kmer_list.append(f">{transcript_name}.kmer{kmers_analysed}\n{mer}")
                    ### contigs case
                    if i == 1:
                        contig_seq = mer
                        position_kmer_prev = position_kmer
                    elif i > 1 and position_kmer == position_kmer_prev + 1:
                        contig_seq = f"{contig_seq}{mer[-1]}"
                        position_kmer_prev = position_kmer
                    else:
                        fasta_contig_list.append(f">{transcript_name}.contig{j}\n{contig_seq}")
                        j += 1
                        contig_seq = mer
                        position_kmer_prev = position_kmer

            ### Case of chimera
            elif level == 'chimera':
                if transcriptome_count == genome_count == 0:
                    ### kmers case
                    i += 1
                    fasta_kmer_list.append(f">{gene_name}.kmer{kmers_analysed}\n{mer}")
                    ### contig case
                    if i == 1:
                        contig_seq = mer
                        position_kmer_prev = position_kmer
                    elif i > 1 and position_kmer == position_kmer_prev + 1:
                        contig_seq = f"{contig_seq}{mer[-1]}"
                        position_kmer_prev = position_kmer
                    else:
                        fasta_contig_list.append(f">{gene_name}.contig{j}\n{contig_seq}")
                        j += 1
                        contig_seq = mer
                        position_kmer_prev = position_kmer
            else:
                sys.exit(f"{Color.RED}Error: level {level} unknown")

        ### append last contig in list
        if level == "gene" and contig_seq:
            fasta_contig_list.append(f">{gene_name}:{transcript_name}.contig{j}\n{contig_seq}")
        elif level == "transcript" and self.args.selection and contig_seq:
            fasta_contig_list.append(f">{given_name}.contig{j}\n{contig_seq}")
        elif (level == "chimera" or (level == "transcript" and self.args.fasta_file)) and contig_seq:
            fasta_contig_list.append(f">{gene_name}.contig{j}\n{contig_seq}")

        ## write tag files
        if fasta_kmer_list:
            tags_outdir = os.path.join(self.args.output, 'tags')
            os.makedirs(tags_outdir, exist_ok=True)
            with open(os.path.join(tags_outdir, tag_file), 'w') as fh:
                fh.write("\n".join(fasta_kmer_list) + '\n')
        else:
            if self.args.selection:
                mesg = (f"{given_name} has no specific kmers found.")
            else:
                mesg = (f"{transcript} has no specific kmers found.")
            return ('aborted', mesg)
        ## write contig files
        if fasta_contig_list:
            contigs_outdir = os.path.join(self.args.output, 'contigs')
            os.makedirs(contigs_outdir, exist_ok=True)
            with open(os.path.join(contigs_outdir, contig_file), 'w') as fh:
                fh.write("\n".join(fasta_contig_list) + '\n')

        if self.args.selection:
            mesg = (f"{gene_name}:{transcript_name} as {level} level ({given_name}).")
        else:
            mesg = (f"{gene_name} as {level} level.")
        return ('done', mesg)


    ### Jellyfish on genome and transcriptome
    def jellyfish(self):
        args = self.args
        genome = args.genome
        ### To create jellyfish PATH DIR
        index_dir = f"{args.output}/indexes"
        mk_jfdir = lambda x: os.makedirs(x, exist_ok=True)

        ### building kmercounts dictionary from jellyfish query on the genome

        ### Compute jellyfish on TRANSCRIPTOME
        if args.debug: print(f"{'-'*9}\n{Color.YELLOW}Compute Jellyfish on the transcriptome.{Color.END}")
        root_path = '.'.join(args.transcriptome.split('.')[:-1])
        root_basename = os.path.basename(root_path)
        jelly_candidate = f"{root_path}.jf"
        jelly_dest = f"{index_dir}/{root_basename}.jf"
        ### check for existing jellyfish transcriptome
        if not args.jellyfish_transcriptome:
            ### at the same location of fasta transcriptome
            if os.path.isfile(jelly_candidate):
                args.jellyfish_transcriptome = jelly_candidate
            ### where the jellyfich file must be created
            if os.path.isfile(jelly_dest):
                args.jellyfish_transcriptome = jelly_dest
        ### do jellyfish on transcriptome fasta file
        if not args.jellyfish_transcriptome:
            tr_root_file = '.'.join(os.path.basename(args.transcriptome).split('.')[:-1])
            args.jellyfish_transcriptome = f"{index_dir}/{tr_root_file}.jf"
            print(" 🧬 Compute Jellyfish on the transcriptome, please wait...")
            mk_jfdir(index_dir)
            cmd = (f"jellyfish count -m {args.kmer_length} -s 1000 -t {args.procs}"
                   f" -o {args.jellyfish_transcriptome} {args.transcriptome}")
            try:
                subprocess.run(cmd, shell=True, check=True, capture_output=True)
            except subprocess.CalledProcessError:
                sys.exit(f"{Color.RED}An error occured in jellyfish count command:\n"
                         f"{cmd}{Color.END}")

        ### Compute jellyfish on GENOME if genome is fasta file
        ext = args.genome.split('.')[-1]
        if ext == "fa" or ext == "fasta":
            mk_jfdir(index_dir)
            indexed_genome = '.'.join(os.path.basename(args.genome).split('.')[:-1]) + '.jf'
            args.jellyfish_genome = os.path.join(index_dir, indexed_genome)
            if os.path.exists(args.jellyfish_genome):
                if args.debug:
                    print(f"{Color.YELLOW}{args.jellyfish_genome} already exists, "
                    f"keep it (manually remove to update it).{Color.END}")
            else:
                print(" 🧬 Compute Jellyfish on the genome, please wait...")
                cmd = (f"jellyfish count -m {args.kmer_length} -s 1000 -t {args.procs}"
                    f" -o {args.jellyfish_genome} {args.genome}")
                try:
                    subprocess.run(cmd, shell=True, check=True, capture_output=True)
                except subprocess.CalledProcessError:
                    sys.exit(f"{Color.RED}An error occured in jellyfish command:\n"
                            f"{cmd}{Color.END}")
        ### When jellyfish genome already exists
        else:
            if args.debug: print(f"{Color.YELLOW}Jellyfish genome index already provided.{Color.END}")
            args.jellyfish_genome = genome

        ### Ending
        if args.debug:
            print(f"{Color.YELLOW}Transcriptome kmer index output: {args.jellyfish_transcriptome}\n"
                  f"Jellyfish done.{Color.END}")






# ~ import sys
import os
import info
import subprocess
from configparser import ConfigParser



DEFAULT_CONFIG = """[CMD_ARGS]

## --genome option
## Path to genome fasta file or jellyfish index of genome
## if fasta file is given, kmerator will create jellyfish index
## We advise you to create a jellyfish index beforehand
# genome = /index/jellyfish/GRCh38_with_MT.jf

## --transcriptome option
## Path to transcriptome (fasta file)
# transcriptome = /transcriptome/GRCh38.fa

## --jellyfish-transcriptome option
## path to jellyfish transcriptome.
## If a jellyfish transcriptome with the same name (exept extension) exists
## in the directory of the --transcriptome option, this parameter is optional.
## If this option is not set and not jellyfish transcriptome found, it will
## be created.
# jellyfish_transcriptome = /transcriptome/GRCh38.jf

## --specie option
## set a specie (default: human)
# specie = human

## --kmer-length option
## set kmer length (default: 31)
# kmer_length = 31

## --chimera
## Only with --fasta-file option
# chimera = False

## --stringent option
## Only with --selection option
stringent = False

## output directory
# outdir = ./output

## --procs option
## number of procs (default: 1)
# procs = 1

## --keep option
## keep intermediate files (default: False)
# keep = False
"""


class Config:
    """
    attributes:
        - filename: name of configuration file
        - filepath: path of the configuration file
        - content: dict of configuration file
    methods:
        - edit(): edit configuration file
    """
    filename = 'config.ini'

    def __init__(self, appname):
        """ Class initialiser """
        ### define location of config file
        self.filePath = self.get_configPath(appname)
        ### create config file if not exists
        if not os.path.isfile(self.filePath):
            self.__set_default()
        ### create a dict of config file
        self.content = ConfigParser()
        self.content.read(self.filePath)


    def get_configPath(self, appname):
        """
        Find to config
            - app_configfile: params of app
            - user_configfile: params for the user (HOTSPOTS, SNP, etc)
        """
        ### Define config file location
        if os.geteuid():
            configdir = os.path.join(os.path.expanduser('~'), '.config', appname)
        else:
            configdir = os.path.join('/etc', appname)
        configfile = os.path.join(configdir, self.filename)
        return configfile


    def __set_default(self):
        try:
            os.makedirs(os.path.dirname(self.filePath), exist_ok=True)
            with open(self.filePath, 'w') as fs:
                fs.write(DEFAULT_CONFIG)
        except PermissionError as err:
            sys.exit(err.msg)


    def edit(self):
        """ Function doc """
        cmd = 'xdg-open' if os.environ.get('XDG_SESSION_TYPE') == 'x11' else 'editor'
        subprocess.call([cmd, self.filePath])



#!/usr/bin/env python3

"""
tools for Kmerator

- mk-transcripts: build a transcriptome and his jellyfish index for a specie
- mk-genome: build a genome and his jellyfish index for a specie --- COMING SOON
- list-species: list all species available from Ensembl API --- COMING SOON
"""

"""
TODO LIST
- [ ] ADD: mk-genome to create genome and jellyfish of genome
- [ ] ADD: list-species to show all species available from Ensembl API
- [ ] ADD: more species (find if specie is available using ebl-species.py)
- [ ] ADD: mk-transcripts: for current release, use date or size or both to not update if it's seems the same
- [ ] ADD: Handle Ctrl + C
- [ ] ADD: Handle alternative to jellyfish, like kmc and kmtricks
"""


import os
import sys
import argparse
import requests
from bs4 import BeautifulSoup
from urllib.request import urlretrieve
import gzip
# import fileinput      # if
import shutil

import info


__appname__   = "ktools"
__shortdesc__ = "tools for kmerator"
__licence__   = info.LICENCE
__version__   = info.VERSION
__author__    = info.AUTHOR



SPECIES = {
    'human': 'homo_sapiens',
    'mouse': 'mus_musculus',
    'zebrafish': 'danio_rerio'
}



def main():
    """ Function doc """
    args = usage()

    if args.command == 'mk-transcripts':
        MkTranscripts(args)


class MkTranscripts:
    """ Class doc """

    def __init__(self, args):
        """ Class initialiser """

        ### select fasta file to download
        base_url = "http://ftp.ensembl.org/pub/"
        release = f"release-{args.release}/fasta" if args.release else "current_fasta"
        specie = SPECIES[args.specie]

        downloaded_fasta = []
        temp_fasta_files = []

        print(f"{__appname__} {args.command} proceed, please wait...")
        ### download fasta gzipped files
        for item in ('cdna', 'ncrna'):

            ### get url links for cDNA and ncRNA fasta files
            url = os.path.join(base_url, release, specie, item)
            file_name = self.get_link(url, item)
            link = os.path.join(url, file_name)

            ### Download fasta files
            fasta_path = self.wget_fasta(args, link, file_name)
            downloaded_fasta.append(fasta_path)

            ### check fasta file
            self.check_fasta(fasta_path)

            ### create a temp file, with alternate chromosome removed
            temp_fasta_path = self.filtered_fasta(args, fasta_path)
            temp_fasta_files.append(temp_fasta_path)

        ### concatene filtered cDNA and ncRNA (and remove temp fasta files)
        transcriptome_fa = self.mk_transcriptome(temp_fasta_files)

        ### make index of transcriptome (kmc or jellyfish)
        self.mk_index(args, transcriptome_fa)

        ### quit properly
        self.exit_gracefully(args, downloaded_fasta, temp_fasta_files)


    def get_link(self, base_url, item):
        if item == 'cdna':
            pattern = 'cdna.all.fa.gz'
        elif item == 'ncrna':
            pattern = 'ncrna.fa.gz'
        try:
            response = requests.get(base_url, timeout=10)
        except requests.exceptions.ConnectionError:
            sys.exit("Error: unable to connect to Ensembl API.")
        if response.ok:
            response_text = response.text
        else:
            return response.raise_for_status()
        soup = BeautifulSoup(response_text, 'html.parser')
        href = [link.get('href') for link in soup.find_all('a') if pattern in link.get('href')][0]
        return href


    def wget_fasta(self, args, link, file_name):
        release = args.release or 'current'
        gz_file = file_name.split('.')
        gz_file = f"{'.'.join(gz_file[0:2])}.{release}.{'.'.join(gz_file[2:])}"
        # ~ print(link)
        fasta_path = os.path.join(args.output, gz_file)
        ### check if Ensembl fasta files already exists
        if os.path.isfile(fasta_path):
            print(f"Notice: {gz_file!r} already exists, it will no be downloaded.")
            return fasta_path
        os.makedirs(args.output, exist_ok=True)
        urlretrieve(link, fasta_path)
        return fasta_path


    def check_fasta(self, fasta_file):
        try :
            with gzip.open(fasta_file) as fh:
                first_line = fh.readline().rstrip().decode()
                if not first_line.startswith(">"):
                    sys.exit("{}Error: Are you sure {} is a fasta file ?{}".format(bcolors.RED, fasta, bcolors.END))
                return first_line
        except FileNotFoundError:
            sys.exit("{}Error: File '{}' not found.{}".format(bcolors.RED, fasta, bcolors.END))


    def filtered_fasta(self, args, fasta_file_path):
        ## Handle sequences
        sequences = []
        seq = []
        current_header = ""
        end_head =  '\n'            # '\t' if args.tsv else '\n'
        sep_seq  =  '\n'            # '' if args.uniq or args.tsv else '\n'
        with gzip.open(fasta_file_path, 'rt') as fh:
            previous_header = fh.readline()
            for line in fh:
                if line.rstrip():
                    if line.startswith('>'):
                        current_header = line
                        if not 'CHR_' in previous_header.split()[2]:
                            sequences.append(f"{previous_header}{''.join(seq)}\n")
                        seq = []
                        previous_header = current_header
                    else:
                        seq.append(line.rstrip())
        ### last fasta sequence is not printed in the loop
        if not 'CHR_' in previous_header.split()[2]:
            sequences.append(f"{previous_header}{''.join(seq)}\n")
        ### write temp file
        fasta_file_basename = ".".join(os.path.basename(fasta_file_path).split('.')[:-2])
        temp_fasta_path = os.path.join(args.output, f'{fasta_file_basename}-altCHR.fa')
        with open(temp_fasta_path, 'w') as fh:
            for sequence in sequences:
                fh.write(sequence)
        ### check if temparry file done
        if not os.path.isfile(temp_fasta_path):
            sys.exit(f"Error: temporary {temp_fasta_path!r} not found.")
        return temp_fasta_path


    def output_files_exists(self, args, temp_fasta_files):
        """ Function doc """
        if not os.isdir(args.output):
            return False
        release = args.release or 'current'
        fasta = SPECIE[args.specie]
        transcriptome_fa = f"{'.'.join(temp_fasta_files[0].split('.')[:3])}.cdna+ncrna-altchr.fa"
        if os.path.isfile(transcriptome_fa):
            print(f"Notice: {transcriptome_fa!r} already exists, it will no be downloaded.")
            return True
        return False


    def mk_transcriptome(self, temp_fasta_files):
        ### check if transcriptom already exists
        transcriptome_fa = f"{'.'.join(temp_fasta_files[0].split('.')[:3])}.cdna+ncrna-altchr.fa"
        if os.path.isfile(transcriptome_fa):
            print(f"Notice: {transcriptome_fa!r} already exists, it will be overwritten.")

        ### concatene cDNA and ncRNA
        print(f"creating transcriptome {os.path.basename(transcriptome_fa)!r}")
        with open(transcriptome_fa, 'w') as outfile:
            for fasta in temp_fasta_files:
                with open(fasta, 'r') as infile:
                    # ~ outfile.write(infile.read())        # small files
                    for line in infile:                     # big files
                        outfile.write(line)
        ''' other method
        with open(transcriptome_fa, 'w') as fout, fileinput.input(temp_fasta_files, 'rb') as fin:
            for line in fin:
                fout.write(line)
        '''
        return transcriptome_fa


    def mk_index(self, args, transcriptome_fa):
        """ Function doc """
        ### Select best tool to indexing transcriptome
        tools = ['_kmc', 'jellyfish']
        num = None
        for i,bin in enumerate(tools):
            if shutil.which(bin):
                num = i
                break
        if not num:
            print(f"Warning: no tool found to index the transcriptome")
            return None

        ### Define command line according to the tool used
        ''' in the future (python 3.10)
        match tools[num]:
            case 'kmc':
                print(f"prefered tool: {os.path.basename(tool)}")
            case 'jellyfish':
                cmd = f"{tool} count -m args.kmer_length -s 100000 {transcriptome_fa} -o {transcriptome_idx}"
        '''
        tool = tools[num]
        ## kmc case
        if tool == 'kmc': print(f"prefered tool: {os.path.basename(tool)}")
        ## jellyfish case
        elif tool == 'jellyfish':
            transcriptome_idx = f"{os.path.splitext(transcriptome_fa)[0]}.jf"
            cmd = f"{tool} count -t {args.thread} -m {args.kmer_length} -s 100000 {transcriptome_fa} -o {transcriptome_idx}"

        ### Build index
        print(f"Build index with {tool!r}, please wait...")
        ret = os.system(cmd)


    def exit_gracefully(self, args, downloaded_fasta, temp_fasta_files):
        ### remove temp fasta files
        if not args.keep:
            for file in downloaded_fasta:
                os.remove(file)
            for file in temp_fasta_files:
                os.remove(file)


def usage():
    """
    Help function with argument parser.
    https://docs.python.org/3/howto/argparse.html?highlight=argparse
    """
    doc_sep = '=' * min(55, os.get_terminal_size(2)[0])
    parser = argparse.ArgumentParser(description= f'{doc_sep}{__doc__}{doc_sep}',
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     prog=__appname__,)
    global_parser = argparse.ArgumentParser(add_help=False)
    subcmd_parser = argparse.ArgumentParser(add_help=False)
    subparsers = parser.add_subparsers()
    subparsers.dest = 'command'

    ### OPTION
    global_parser.add_argument('-o', '--output',
                        help = "output directory (default: .)",
                        default=".",
                       )
    parser_mk_tcrptome = subparsers.add_parser('mk-transcripts',
                        parents = [global_parser],
                        help="make transcriptome",
                        )
    parser_mk_tcrptome.add_argument('-s', '--specie',
                        type=str,
                        help="human or mouse (default: human)",
                        default="human",
                        choices=SPECIES,
                        )
    parser_mk_tcrptome.add_argument('-r', '--release',
                        help="Encode version of transcriptome to dowload (default: current)",
                        )
    parser_mk_tcrptome.add_argument('-t', '--thread',
                        type=int,
                        help="number of trhead (defaul: 1)",
                        default=1,
                       )
    parser_mk_tcrptome.add_argument('-k', '--kmer-length',
                        type=int,
                        help="length of kmer (defaul: 31)",
                        default=31,
                       )
    parser_mk_tcrptome.add_argument('--keep',
                        action="store_true",          # boolean
                        help="Keep intermediates files",
                       )
    parser.add_argument('-v', '--version',
                        action='version',
                        version=f"{parser.prog} v{__version__}",
                       )
    ### Go to "usage()" without arguments or stdin
    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit()
    return parser.parse_args()


if __name__ == "__main__":
    main()



### fasta.py

import os
import sys
from utils import Color


def fasta2dict(fasta_file):
    '''
    Basic convertion of fasta file to dict
    It keeps all the header
    '''
    ### controls
    with open(fasta_file) as fh:
        first_line = fh.readline()
        if not first_line.startswith('>'):
            sys.exit(f"{Color.RED}Error: {os.path.basename(fasta_file)!r} does not seem to be in fasta format.")
    ### compute file as dict
    fasta_dict = {}
    with open(fasta_file) as fh:
        seq = ""
        old_desc, new_desc = "", ""
        for line in fh:
            if line[0] == ">":
                new_desc = line.rstrip().lstrip('>')
                if old_desc:
                    fasta_dict[old_desc] = seq
                    seq = ""
                old_desc = new_desc
            else:
                seq += line.upper().rstrip()
    fasta_dict[old_desc] = seq
    return fasta_dict



### sequences.py


import os
import sys
from concurrent.futures import ThreadPoolExecutor

from utils import Color
from fasta import fasta2dict


class Sequences:
    """ Class doc """

    def __init__(self, args, report, transcripts, transcriptome_dict=None):
        """ Class initialiser """
        self.args = args
        self.report = report
        self.transcripts = transcripts
        self.transcriptome_dict = transcriptome_dict
        self.removed_transcripts = [] # list of transcript to remove when not found in transcriptome


    def build(self):
        """
        create files for each transcript
        Notice:
            transcripts == best_transcripts  when genes/transcripts are known
            transcripts == unannotated_transcripts  for unannotated sequences
        """
        output_seq_dir = os.path.join(self.args.output, 'sequences')
        ### Whith --selection option
        if self.args.selection:
            ### Abort if no transcripts found
            if not self.transcripts:
                sys.exit(f"{Color.RED}\n Error: no transcript found for {self.args.selection}{Color.END}\n")
            ### create output directory structure
            os.makedirs(output_seq_dir, exist_ok=True)

            ### create fasta files for each gene/transcript of selection found in transcriptome
            transcripts_list = [ {**{'enst':k}, **v} for k,v in self.transcripts.items()]
            with ThreadPoolExecutor(self.args.procs) as pool:
                list(pool.map(self.mk_annot_seq, transcripts_list))
            for trscrpt in self.removed_transcripts:
                self.transcripts.pop(trscrpt)
        ### With --fasta-file option
        else:
            ### read fasta file
            if self.args.debug: print(f"{Color.YELLOW}{'-'*12}\n\nBuild sequences without transcriptome.\n{Color.END}")
            fastafile_dict = fasta2dict(self.args.fasta_file)
            ### Abort if dict empy
            if not fastafile_dict:
                sys.exit(f"{Color.RED}Error: no sequence found for {args.fasta_file}")
            ### create output directory structure
            os.makedirs(output_seq_dir, exist_ok=True)
            ### create fasta files for unannotated sequences
            fastafile_list = [ {'desc': k, 'seq':v} for k,v in fastafile_dict.items()]
            with ThreadPoolExecutor(self.args.procs) as pool:
                list(pool.map(self.mk_unannot_seq, fastafile_list))


    def mk_annot_seq(self, transcript):
        """
        find canonical transcript in transcriptome and write his sequence in file.
        """
        # ~ if not 'symbol' in transcript:
            # ~ print(f"{Color.RED}No symbol name found ({transcript}.{Color.END}")
        desc = f"{transcript['symbol']}:{transcript['enst']}"

        if desc in self.transcriptome_dict:
            seq = self.transcriptome_dict[desc]
            if len(seq) < self.args.kmer_length:
                self.report['aborted'].append(f"{transcript['given']!r} has a sequence length < {self.args.kmer_length} => ignored")
                self.removed_transcripts.append(transcript['enst'])
                return
            ### create fasta files
            outfile = f"{transcript['symbol']}.{transcript['enst']}.fa" # [:255].replace(' ', '_').replace('/', '@SLASH@')
            outfile = f"{self.args.output}/sequences/{outfile}"
            with open(outfile, 'w') as fh:
                fh.write(f">{transcript['symbol']}:{transcript['enst']}\n{seq}")
            ### check if file is present
            if not os.path.isfile(outfile):
                print(f"{Color.RED}Error: unable to create {outfile}.")
        ### When transcript is not found
        else:
            self.report['aborted'].append(f"{transcript['enst']!r} not found in provided transcriptome (gene: {transcript['symbol']})")
            self.removed_transcripts.append(transcript['enst'])


    def mk_unannot_seq(self, entry):
        """ Function doc """
        outfile = f"{entry['desc'].replace(' ', '_').replace('/', '@SLASH@')}.fa"[:255]
        outfile = os.path.join(self.args.output, 'sequences', outfile)
        if len(entry['seq']) < self.args.kmer_length:
            self.report['aborted'].append(f"{entry['desc']} sequence length < {self.args.kmer_length} => ignored.")
            self.removed_transcripts.append(entry['desc'])
            return
        self.transcripts.append(entry['desc'])
        with open(outfile, 'w') as fh:
            fh.write(f">{entry['desc'][:79]}\n{entry['seq']}")
        ### check if file is present
        if not os.path.isfile(outfile):
            print(f"{Color.RED}Error: unable to create {outfile}.")
